﻿// DataTypes.cs

using System;

// if we're building the MSRS version, we need to bring in the MSRS Attributes
// if we're not doing the MSRS build then define some fake attribute classes for DataMember/DataContract
#if MSRS
	using Microsoft.Dss.Core.Attributes;
#else
sealed class DataContract : Attribute
{
}

sealed class DataMember : Attribute
{
}
#endif


namespace WaveLib
{
#if MSRS
    [DataContract]
    public struct RumbleRequest
    {
        [DataMember]
        public bool Rumble;
    }
#endif

    /// <summary>
    /// Point structure for floating point 2D positions (X, Y)
    /// </summary>
    [Serializable]
    [DataContract]
    public struct PointF
    {
        /// <summary>
        /// X, Y coordinates of this point
        /// </summary>
        [DataMember]
        public float X, Y;

        /// <summary>
        /// Convert to human-readable string
        /// </summary>
        /// <returns>A string that represents the point</returns>
        public override string ToString()
        {
            return string.Format("{{X={0}, Y={1}}}", X, Y);
        }

    }

    /// <summary>
    /// Point structure for int 2D positions (X, Y)
    /// </summary>
    [Serializable]
    [DataContract]
    public struct Point
    {
        /// <summary>
        /// X, Y coordinates of this point
        /// </summary>
        [DataMember]
        public int X, Y;

        /// <summary>
        /// Convert to human-readable string
        /// </summary>
        /// <returns>A string that represents the point.</returns>
        public override string ToString()
        {
            return string.Format("{{X={0}, Y={1}}}", X, Y);
        }
    }

    /// <summary>
    /// Point structure for floating point 3D positions (X, Y, Z)
    /// </summary>
    [Serializable]
    [DataContract]
    public struct Point3F
    {
        /// <summary>
        /// X, Y, Z coordinates of this point
        /// </summary>
        [DataMember]
        public float X, Y, Z;

        /// <summary>
        /// Convert to human-readable string
        /// </summary>
        /// <returns>A string that represents the point</returns>
        public override string ToString()
        {
            return string.Format("{{X={0}, Y={1}, Z={2}}}", X, Y, Z);
        }

    }

    /// <summary>
    /// Point structure for int 3D positions (X, Y, Z)
    /// </summary>
    [Serializable]
    [DataContract]
    public struct Point3
    {
        /// <summary>
        /// X, Y, Z coordinates of this point
        /// </summary>
        [DataMember]
        public int X, Y, Z;

        /// <summary>
        /// Convert to human-readable string
        /// </summary>
        /// <returns>A string that represents the point.</returns>
        public override string ToString()
        {
            return string.Format("{{X={0}, Y={1}, Z={2}}}", X, Y, Z);
        }
    }

    /// <summary>
    /// Current overall state of the Wave
    /// </summary>
    [Serializable]
    [DataContract]
    public class WaveState
    {
        /// <summary>
        /// Current state of buttons
        /// </summary>
        [DataMember]
        public ButtonState ButtonState;
        /// <summary>
        /// Current state of encoders
        /// </summary>
        [DataMember]
        public EncoderState EncoderState;
        /// <summary>
        /// Current state of firmware
        /// </summary>
        [DataMember]
        public FirmwareState FirmwareState;
        /// <summary>
        /// Constructor for WaveState class
        /// </summary>
        public WaveState()
        {
        }
    }

    /// <summary>
    /// Current button state
    /// </summary>
    [Serializable]
    [DataContract]
    public struct ButtonState
    {
        /// <summary>
        /// Digital button on the Wiimote
        /// </summary>
        [DataMember]
        public bool Up, Down, Enc1, Enc2, Enc3, Enc4, Enc5, Enc6, Enc7, Enc8, Enc9, B1, B2, B3, B4, B5, B6, B7, B8, B9, Alt, REnc1, RTr1, REnc2, RTr2, REnc3, RTr3, F1, F2, F3, F4, F5, F6, F7, F8, F9, Reverse, Play, Stop, Previous, Next;
    }

    /// <summary>
    /// Current state of the encoders
    /// </summary>
    [Serializable]
    [DataContract]
    public struct EncoderState
    {
        /// <summary>
        /// Encoder data.
        /// <remarks>Values range between -127 - 127</remarks>
        /// </summary>
        [DataMember]
        public short Enc1, Enc2, Enc3, Enc4, Enc5, Enc6, Enc7, Enc8, Enc9, Enc10, Enc11, Enc12, Enc13;
        /// <summary>
        /// TrackBall data.
        /// <remarks>Values range between -127 - 127</remarks>
        /// </summary>
        [DataMember]
        public Point Tr1, Tr2, Tr3;
    }

    /// <summary>
    /// Current state of the firmware
    /// </summary>
    [Serializable]
    [DataContract]
    public struct FirmwareState
    {
        /// <summary>
        /// Firmware data.
        /// <remarks>Reserved for use during firmware updates</remarks>
        /// </summary>
        [DataMember]
        public Byte Firmware;
    }

    /// <summary>
    /// Last ReadData status
    /// </summary>
    public enum LastReadStatus
    {
        /// <summary>
        /// Successful read
        /// </summary>
        Success,
        /// <summary>
        /// Attempt to read from write only memory
        /// </summary>
        ReadFromWriteOnlyMemory
    };

}
