﻿// Wave.cs

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Runtime.InteropServices;
using System.Diagnostics;
using System.IO;
using System.Runtime.Serialization;
using Microsoft.Win32.SafeHandles;
using System.Threading;

namespace WaveLib
{
    /// <summary>
    /// Implementation of Wave
    /// </summary>
    public class Wave : IDisposable
    {
        /// <summary>
        /// Event raised when Wave state is changed
        /// </summary>
        public event EventHandler<WaveChangedEventArgs> WaveChanged;

        // VID = TangentDevices, PID = Wave
        private const ushort VID = 0x04d8;
        private const ushort PID = 0xfdcf;

        // sure, we could find this out the hard way using HID, but trust me, it's 41
        private const int REPORT_LENGTH = 41;

        // read/write handle to the device
        private SafeFileHandle mHandle;

        // a pretty .NET stream to read/write from/to
        private FileStream mStream;

        // current state of controller
        private readonly WaveState mWaveState = new WaveState();

        // event for read data processing
        private readonly AutoResetEvent mReadDone = new AutoResetEvent(false);
        private readonly AutoResetEvent mWriteDone = new AutoResetEvent(false);

        // event for status report
        private readonly AutoResetEvent mStatusDone = new AutoResetEvent(false);

        // use a different method to write reports
        private bool mAltWriteMethod;

        // HID device path of this Wiimote
        private string mDevicePath = string.Empty;

        // unique ID
        private readonly Guid mID = Guid.NewGuid();

        // delegate used for enumerating found Waves
        internal delegate bool WaveFoundDelegate(string devicePath);

        /// <summary>
		/// Default constructor
		/// </summary>
		public Wave()
		{
		}

		internal Wave(string devicePath)
		{
			mDevicePath = devicePath;
		}

        /// <summary>
        /// Connect to the first-found Wave
        /// </summary>
        /// <exception cref="WaveNotFoundException">Wave not found in HID device list</exception>
        public void Connect()
        {
            if (string.IsNullOrEmpty(mDevicePath))
                FindWave(WaveFound);
            else
                OpenWaveDeviceHandle(mDevicePath);
        }

        internal static void FindWave(WaveFoundDelegate waveFound)
        {
            int index = 0;
            bool found = false;
            Guid guid;
            SafeFileHandle mHandle;

            // get the GUID of the HID class
            HIDImports.HidD_GetHidGuid(out guid);

            // get a handle to all devices that are part of the HID class
            // Fun fact:  DIGCF_PRESENT worked on my machine just fine.  I reinstalled Vista, and now it no longer finds the Wiimote with that parameter enabled...
            IntPtr hDevInfo = HIDImports.SetupDiGetClassDevs(ref guid, null, IntPtr.Zero, HIDImports.DIGCF_DEVICEINTERFACE);// | HIDImports.DIGCF_PRESENT);

            // create a new interface data struct and initialize its size
            HIDImports.SP_DEVICE_INTERFACE_DATA diData = new HIDImports.SP_DEVICE_INTERFACE_DATA();
            diData.cbSize = Marshal.SizeOf(diData);

            // get a device interface to a single device (enumerate all devices)
            while (HIDImports.SetupDiEnumDeviceInterfaces(hDevInfo, IntPtr.Zero, ref guid, index, ref diData))
            {
                UInt32 size;

                // get the buffer size for this device detail instance (returned in the size parameter)
                HIDImports.SetupDiGetDeviceInterfaceDetail(hDevInfo, ref diData, IntPtr.Zero, 0, out size, IntPtr.Zero);

                // create a detail struct and set its size
                HIDImports.SP_DEVICE_INTERFACE_DETAIL_DATA diDetail = new HIDImports.SP_DEVICE_INTERFACE_DETAIL_DATA();

                // yeah, yeah...well, see, on Win x86, cbSize must be 5 for some reason.  On x64, apparently 8 is what it wants.
                // someday I should figure this out.  Thanks to Paul Miller on this...
                diDetail.cbSize = (uint)(IntPtr.Size == 8 ? 8 : 5);

                // actually get the detail struct
                if (HIDImports.SetupDiGetDeviceInterfaceDetail(hDevInfo, ref diData, ref diDetail, size, out size, IntPtr.Zero))
                {
                    Debug.WriteLine(string.Format("{0}: {1} - {2}", index, diDetail.DevicePath, Marshal.GetLastWin32Error()));

                    // open a read/write handle to our device using the DevicePath returned
                    mHandle = HIDImports.CreateFile(diDetail.DevicePath, FileAccess.ReadWrite, FileShare.ReadWrite, IntPtr.Zero, FileMode.Open, HIDImports.EFileAttributes.Overlapped, IntPtr.Zero);

                    // create an attributes struct and initialize the size
                    HIDImports.HIDD_ATTRIBUTES attrib = new HIDImports.HIDD_ATTRIBUTES();
                    attrib.Size = Marshal.SizeOf(attrib);

                    // get the attributes of the current device
                    if (HIDImports.HidD_GetAttributes(mHandle.DangerousGetHandle(), ref attrib))
                    {
                        // if the vendor and product IDs match up
                        if (attrib.VendorID == VID && (ushort)attrib.ProductID == PID)
                        {
                            // it's a Wave
                            Debug.WriteLine("Found one!");
                            found = true;

                            // fire the callback function...if the callee doesn't care about more Waves, break out
                            if (!waveFound(diDetail.DevicePath))
                                break;
                        }
                    }
                    mHandle.Close();
                }
                else
                {
                    // failed to get the detail struct
                    throw new WaveException("SetupDiGetDeviceInterfaceDetail failed on index " + index);
                }

                // move to the next device
                index++;
            }

            // clean up our list
            HIDImports.SetupDiDestroyDeviceInfoList(hDevInfo);

            // if we didn't find a Wiimote, throw an exception
            if (!found)
                throw new WaveNotFoundException("No Wave found in HID device list.");
        }

        private bool WaveFound(string devicePath)
        {
            mDevicePath = devicePath;

            // if we didn't find a Wave, throw an exception
            OpenWaveDeviceHandle(mDevicePath);

            return false;
        }

        private void OpenWaveDeviceHandle(string devicePath)
        {
            // open a read/write handle to our device using the DevicePath returned
            mHandle = HIDImports.CreateFile(devicePath, FileAccess.ReadWrite, FileShare.ReadWrite, IntPtr.Zero, FileMode.Open, HIDImports.EFileAttributes.Overlapped, IntPtr.Zero);

            // create an attributes struct and initialize the size
            HIDImports.HIDD_ATTRIBUTES attrib = new HIDImports.HIDD_ATTRIBUTES();
            attrib.Size = Marshal.SizeOf(attrib);

            // get the attributes of the current device
            if (HIDImports.HidD_GetAttributes(mHandle.DangerousGetHandle(), ref attrib))
            {
                // if the vendor and product IDs match up
                if (attrib.VendorID == VID && (ushort)attrib.ProductID == PID)
                {
                    // create a nice .NET FileStream wrapping the handle above
                    mStream = new FileStream(mHandle, FileAccess.ReadWrite, REPORT_LENGTH, true);

                    // start an async read operation on it
                    BeginAsyncRead();

                    // read the calibration info from the controller
                    try
                    {
                    }
                    catch
                    {
                        // if we fail above, try the alternate HID writes
                        mAltWriteMethod = true;
                    }

                }
                else
                {
                    // otherwise this isn't the controller, so close up the file handle
                    mHandle.Close();
                    throw new WaveException("Attempted to open a non-Wave device.");
                }
            }
        }

        /// <summary>
        /// Disconnect from the controller and stop reading data from it
        /// </summary>
        public void Disconnect()
        {
            // close up the stream and handle
            if (mStream != null)
                mStream.Close();

            if (mHandle != null)
                mHandle.Close();
        }

        /// <summary>
        /// Start reading asynchronously from the controller
        /// </summary>
        private void BeginAsyncRead()
        {
            // if the stream is valid and ready
            if (mStream != null && mStream.CanRead)
            {
                // setup the read and the callback
                byte[] buff = new byte[REPORT_LENGTH];
                mStream.BeginRead(buff, 0, REPORT_LENGTH, new AsyncCallback(OnReadData), buff);
            }
        }

        /// <summary>
        /// Callback when data is ready to be processed
        /// </summary>
        /// <param name="ar">State information for the callback</param>
        private void OnReadData(IAsyncResult ar)
        {
            // grab the byte buffer
            byte[] buff = (byte[])ar.AsyncState;

            try
            {
                // end the current read
                mStream.EndRead(ar);

                // parse it
                if (ParseInputReport(buff))
                {
                    // post an event
                    if (WaveChanged != null)
                        WaveChanged(this, new WaveChangedEventArgs(mWaveState));
                }

                // start reading again
                BeginAsyncRead();
            }
            catch (OperationCanceledException)
            {
                Debug.WriteLine("OperationCanceledException");
            }
        }

        /// <summary>
        /// Parse a report sent by the Wave
        /// </summary>
        /// <param name="buff">Data buffer to parse</param>
        /// <returns>Returns a boolean noting whether an event needs to be posted</returns>
        private bool ParseInputReport(byte[] buff)
        {
            ParseButtons(buff);
            ParseEncoders(buff);
            ParseReadData(buff);
            return true;
        }

        /*
        /// DEL ?
        /// <summary>
        /// Decrypts data sent from the extension to the Wave
        /// </summary>
        /// <param name="buff">Data buffer</param>
        /// <returns>Byte array containing decoded data</returns>
        private byte[] DecryptBuffer(byte[] buff)
        {
            for (int i = 0; i < buff.Length; i++)
                buff[i] = (byte)(((buff[i] ^ 0x17) + 0x17) & 0xff);

            return buff;
        }
        */

        /// <summary>
        /// Parses a standard button report into the ButtonState struct
        /// </summary>
        /// <param name="buff">Data buffer</param>
        private void ParseButtons(byte[] buff)
        {
            mWaveState.ButtonState.Enc1 = (buff[1] & 0x01) != 0;
            mWaveState.ButtonState.Enc2 = (buff[1] & 0x02) != 0;
            mWaveState.ButtonState.Enc3 = (buff[1] & 0x04) != 0;
            mWaveState.ButtonState.Enc4 = (buff[1] & 0x08) != 0;
            mWaveState.ButtonState.Enc5 = (buff[1] & 0x10) != 0;
            mWaveState.ButtonState.Enc6 = (buff[1] & 0x20) != 0;
            mWaveState.ButtonState.Enc7 = (buff[1] & 0x40) != 0;
            mWaveState.ButtonState.Enc8 = (buff[1] & 0x80) != 0;
            mWaveState.ButtonState.Enc9 = (buff[2] & 0x01) != 0;
            mWaveState.ButtonState.Alt = (buff[2] & 0x02) != 0;
            mWaveState.ButtonState.B1 = (buff[2] & 0x04) != 0;
            mWaveState.ButtonState.B2 = (buff[2] & 0x08) != 0;
            mWaveState.ButtonState.B3 = (buff[2] & 0x10) != 0;
            mWaveState.ButtonState.REnc1 = (buff[2] & 0x20) != 0;
            mWaveState.ButtonState.RTr1 = (buff[2] & 0x40) != 0;
            mWaveState.ButtonState.B4 = (buff[2] & 0x80) != 0;
            mWaveState.ButtonState.B5 = (buff[3] & 0x01) != 0;
            mWaveState.ButtonState.B6 = (buff[3] & 0x02) != 0;
            mWaveState.ButtonState.REnc2 = (buff[3] & 0x04) != 0;
            mWaveState.ButtonState.RTr2 = (buff[3] & 0x08) != 0;
            mWaveState.ButtonState.B7 = (buff[3] & 0x10) != 0;
            mWaveState.ButtonState.B8 = (buff[3] & 0x20) != 0;
            mWaveState.ButtonState.B9 = (buff[3] & 0x40) != 0;
            mWaveState.ButtonState.REnc3 = (buff[3] & 0x80) != 0;
            mWaveState.ButtonState.RTr3 = (buff[4] & 0x01) != 0;
            mWaveState.ButtonState.Up = (buff[4] & 0x02) != 0;
            mWaveState.ButtonState.Down = (buff[4] & 0x04) != 0;
            mWaveState.ButtonState.F7 = (buff[4] & 0x08) != 0;
            mWaveState.ButtonState.F8 = (buff[4] & 0x10) != 0;
            mWaveState.ButtonState.F9 = (buff[4] & 0x20) != 0;
            mWaveState.ButtonState.F4 = (buff[4] & 0x40) != 0;
            mWaveState.ButtonState.F5 = (buff[4] & 0x80) != 0;
            mWaveState.ButtonState.F6 = (buff[5] & 0x01) != 0;
            mWaveState.ButtonState.F1 = (buff[5] & 0x02) != 0;
            mWaveState.ButtonState.F2 = (buff[5] & 0x04) != 0;
            mWaveState.ButtonState.F3 = (buff[5] & 0x08) != 0;
            mWaveState.ButtonState.Previous = (buff[5] & 0x10) != 0;
            mWaveState.ButtonState.Next = (buff[5] & 0x20) != 0;
            mWaveState.ButtonState.Reverse = (buff[5] & 0x40) != 0;
            mWaveState.ButtonState.Stop = (buff[5] & 0x80) != 0;
            mWaveState.ButtonState.Play = (buff[6] & 0x01) != 0;
        }

        /// <summary>
        /// Parses a standard encoder report into the EncoderState struct
        /// </summary>
        /// <param name="buff">Data buffer</param>
        private void ParseEncoders(byte[] buff)
        {
            mWaveState.EncoderState.Enc1 = (sbyte)buff[07];
            mWaveState.EncoderState.Enc2 = (sbyte)buff[08];
            mWaveState.EncoderState.Enc3 = (sbyte)buff[09];
            mWaveState.EncoderState.Enc4 = (sbyte)buff[10];
            mWaveState.EncoderState.Enc5 = (sbyte)buff[11];
            mWaveState.EncoderState.Enc6 = (sbyte)buff[12];
            mWaveState.EncoderState.Enc7 = (sbyte)buff[13];
            mWaveState.EncoderState.Enc8 = (sbyte)buff[14];
            mWaveState.EncoderState.Enc9 = (sbyte)buff[15];
            mWaveState.EncoderState.Enc10 = (sbyte)buff[16];
            mWaveState.EncoderState.Enc11 = (sbyte)buff[17];
            mWaveState.EncoderState.Enc12 = (sbyte)buff[18];
            mWaveState.EncoderState.Enc13 = (sbyte)buff[19];
            mWaveState.EncoderState.Tr1.X = (sbyte)buff[20];
            mWaveState.EncoderState.Tr1.Y = (sbyte)buff[21];
            mWaveState.EncoderState.Tr2.X = (sbyte)buff[22];
            mWaveState.EncoderState.Tr2.Y = (sbyte)buff[23];
            mWaveState.EncoderState.Tr3.X = (sbyte)buff[24];
            mWaveState.EncoderState.Tr3.Y = (sbyte)buff[25];

        }

        /// <summary>
        /// Parse data returned from a read report
        /// </summary>
        /// <param name="buff">Data buffer</param>
        private void ParseReadData(byte[] buff)
        {
            mWaveState.FirmwareState.Firmware = buff[26];

            LastReadStatus = LastReadStatus.Success;

        }

        /// <summary>
        /// Initialize the report data buffer
        /// </summary>
        private byte[] CreateReport()
        {
            return new byte[REPORT_LENGTH];
        }


        /// <summary>
        /// Write a report to the Wave
        /// </summary>
        private void WriteReport(byte[] buff)
        {
            if (mAltWriteMethod)
                HIDImports.HidD_SetOutputReport(this.mHandle.DangerousGetHandle(), buff, (uint)buff.Length);
            else if (mStream != null)
                mStream.Write(buff, 0, REPORT_LENGTH);
        }

        /// <summary>
        /// Write a byte array to Wave LED
        /// </summary>
        /// <param name="id">dispalyID  0-2</param>
        /// <param name="line">Line Number 0-4</param>
        /// <param name="position">Position 0-31</param>
        /// <param name="length">Length of buffe 1-32</param>
        /// <param name="data">Data buffe  MAX 32byte</param>
        public void WriteLED(byte id, byte line, byte position, byte length, byte[] data)
        {
            byte[] buff = CreateReport();
            length = (byte)data.Length;

            if (length > 33)
            {
                length = 32;
            }

            buff[0] = 0x00; //0x00?
            buff[1] = 0x00; //command
            buff[2] = id; //ID 
            buff[3] = line; //Line
            buff[4] = position; //Posiiotn 
            buff[5] = length; //length
            Array.Copy(data, 0, buff, 6, length);

            WriteReport(buff);
        }

        /// <summary>
        /// Write a byte array to Wave Custom Character
        /// </summary>
        /// <param name="id">character ID  0-32</param>
        /// <param name="data">Data buffe 13byte</param>
        public void WriteCustomCharacter(byte id, byte[] data)
        {
            byte[] buff = new byte[15];

            buff[0] = 0x00; //0x00?
            buff[1] = 0x01; //command
            buff[2] = id; //Character ID 
            Array.Copy(data, 0, buff, 3, 13);

            WriteReport(buff);
        }

        /// <summary>
        /// Current Wave state
        /// </summary>
        public WaveState WaveState
        {
            get { return mWaveState; }
        }

        ///<summary>
        /// Unique identifier for this Wave (not persisted across application instances)
        ///</summary>
        public Guid ID
        {
            get { return mID; }
        }

        /// <summary>
        /// HID device path for this Wave (valid until Wave is disconnected)
        /// </summary>
        public string HIDDevicePath
        {
            get { return mDevicePath; }
        }

        /// <summary>
        /// Status of last ReadMemory operation
        /// </summary>
        public LastReadStatus LastReadStatus { get; private set; }

        #region IDisposable Members

        /// <summary>
        /// Dispose Wave
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Dispose wave
        /// </summary>
        /// <param name="disposing">Disposing?</param>
        protected virtual void Dispose(bool disposing)
        {
            // close up our handles
            if (disposing)
                Disconnect();
        }
        #endregion
    }

    /// <summary>
    /// Thrown when no Waves are found in the HID device list
    /// </summary>
    [Serializable]
    public class WaveNotFoundException : ApplicationException
    {
        /// <summary>
        /// Default constructor
        /// </summary>
        public WaveNotFoundException()
        {
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="message">Error message</param>
        public WaveNotFoundException(string message)
            : base(message)
        {
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="message">Error message</param>
        /// <param name="innerException">Inner exception</param>
        public WaveNotFoundException(string message, Exception innerException)
            : base(message, innerException)
        {
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="info">Serialization info</param>
        /// <param name="context">Streaming context</param>
        protected WaveNotFoundException(SerializationInfo info, StreamingContext context)
            : base(info, context)
        {
        }
    }

    /// <summary>
    /// Represents errors that occur during the execution of the Wave library
    /// </summary>
    [Serializable]
    public class WaveException : ApplicationException
    {
        /// <summary>
        /// Default constructor
        /// </summary>
        public WaveException()
        {
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="message">Error message</param>
        public WaveException(string message)
            : base(message)
        {
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="message">Error message</param>
        /// <param name="innerException">Inner exception</param>
        public WaveException(string message, Exception innerException)
            : base(message, innerException)
        {
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="info">Serialization info</param>
        /// <param name="context">Streaming context</param>
        protected WaveException(SerializationInfo info, StreamingContext context)
            : base(info, context)
        {
        }
    }


}
